#!/usr/bin/env python
# -*- coding: utf-8 -*-
from random import randint, seed
from roller_constant_for import BackdropStyle as bs
from roller_constant_key import Option as ok
from roller_fu import (
    color_fill_selection,
    merge_layer_group,
    select_polygon,
    select_rect
)
from roller_maya_style import Style
from roller_one import make_2d_table, seed_random
from roller_view_preset import combine_seed
from roller_view_real import (
    add_base_layer, add_sub_base_group, add_wip_layer, finish_style
)
import gimpfu as fu

pdb = fu.pdb


def create_background(v, j, z, d, q_x, column, right_x):
    """
    Create stripes made of two colors that will
    provide a background for the stretched boxes.

    v: View
    j: GIMP image
    z: layer
        Receive color column.

    q_x: list
    column: int
        Is the column count from the user interface.

    right_x: int
    """
    for p_x, color in enumerate(d[ok.COLOR_3][1:]):
        pdb.gimp_selection_none(j)

        for c in range(column):
            c1 = c * 2

            if p_x % 2:
                x = q_x[c1 + 1]
                w = q_x[c1 + 2] - x

            else:
                x = q_x[c1]
                w = q_x[c1 + 1] - x
            if x >= v.wip.x - w and x < right_x:
                select_rect(j, x, v.wip.y, w, v.wip.h, option=fu.CHANNEL_OP_ADD)
        color_fill_selection(z, color)


def make_style(v, maya):
    """
    Make a style layer for the Backdrop Style.

    v: View
    maya: Style
    Return: layer or None
        style layer
    """
    def _cell_is_valid(_r, _c):
        """
        The cells are double spaced, and the cell at (0, 0) is not a cell.

        _r, _c: numeric
            cell index

        Return: bool
            Is True if the cell is valid.
        """
        if is_odd and _r == 0 and _c == 0:
            return False
        return bool((_r % 2 and _c % 2) or (not _r % 2 and not _c % 2))

    def _do_left():
        """Select the left face of the box."""
        select_polygon(
            j,
            (
                x, y1,
                x1, y2,
                x1, bottom_y,
                x, bottom_y
            ),
            option=fu.CHANNEL_OP_ADD
        )

    def _do_right():
        """Select the right face of the box."""
        select_polygon(
            j,
            (
                x1, y2,
                x2, y1,
                x2, bottom_y,
                x1, bottom_y
            ),
            option=fu.CHANNEL_OP_ADD
        )

    def _do_top():
        """Select the top face of the box."""
        select_polygon(
            j,
            (
                x, y1,
                x1, y,
                x2, y1,
                x1, y2
            ),
            option=fu.CHANNEL_OP_ADD
        )

    def _get_previous_y():
        """
        Ensure that the boxes in the row above
        the current cell are above the current cell.

        Return: float
            y-coordinate of the limiting box
            at the lowest position in the graph.
        """
        y = v.wip.y - row_h
        _r = r - 1
        _c = c - 1
        _c1 = c + 1

        if _r >= 0:
            if _c >= 0 and _cell_is_valid(_r, _c):
                y = table_y[_r][_c] + top_h
            if _r <= row and _c1 <= column:
                y = max(y, table_y[_r][_c1] + top_h)
        return y

    j = v.j
    d = maya.value_d
    parent = add_sub_base_group(v, maya)

    combine_seed(v, d)

    # Calculate grid.
    row, column = int(d[ok.ROW]), int(d[ok.COLUMN])
    a, b = divmod(column, 2.)
    double_column = a + b
    column_w = round(v.wip.w / double_column)
    row_h = round(v.wip.h / row)
    half_column_w = column_w / 2.
    top_h = min(half_column_w, row_h / 2.)
    bottom_y = v.wip.y + v.wip.h
    canvas_x = v.wip.x - half_column_w
    right_x = v.wip.x + v.wip.w
    is_odd = column % 2
    row1 = row + 1
    column1 = column + 1

    # x intersect coordinate
    q_x = []

    # remainder total, 'f_x'
    f_x = .0

    x = canvas_x

    for c in range(column + 1):
        x, f = divmod(x, 1.)
        f_x += f

        if f_x >= .999:
            x += 1.
            f_x -= 1.

        q_x.append(round(x))
        q_x.append(round(x + half_column_w))
        x += column_w

    table_y = make_2d_table(row1, column1)

    # Limit random y with row space.
    for r in range(row):
        y = v.wip.y
        for c in range(column1):
            if _cell_is_valid(r, c):
                # Randomize height to get y.
                y = _get_previous_y()

                if r > 0:
                    y = float(randint(int(y), int(y + row_h)))

                else:
                    y = float(randint(int(y), int(y + top_h)))
                table_y[r][c] = y

    z = add_base_layer(v, parent, n="Background")

    create_background(v, j, z, d, q_x, column, right_x)

    # Each row has its own layer.
    for r in range(row):
        z = add_wip_layer(v, "Row " + str(r), parent)

        for p_x, color in enumerate(d[ok.COLOR_3]):
            pdb.gimp_selection_none(j)
            for c in range(column1):
                if _cell_is_valid(r, c):
                    # Accumulate selection for color fill.
                    x = q_x[c]
                    x1 = q_x[c + 1]
                    x2 = q_x[c + 2]
                    y = table_y[r][c]
                    if y < bottom_y:
                        y1 = y + top_h
                        y2 = y1 + top_h
                        (_do_top, _do_left, _do_right)[p_x]()
            color_fill_selection(z, color)
    return finish_style(merge_layer_group(parent), "Roof Top")


class RoofTop(Style):
    """Create Backdrop Style output."""
    is_dependent = False
    is_seeded = True

    def __init__(self, any_group, super_maya, k_path):
        k_path = [k_path, k_path + (ok.BRW,)]
        Style.__init__(self, any_group, super_maya, k_path, make_style)
