#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_for import Mask as ms, Shape as sh, Triangle as ft
from roller_constant_key import Option as ok
from roller_fu import (
    add_layer,
    color_fill_selection,
    copy_all_image,
    discard_mask,
    get_select_coord,
    get_select_bounds,
    load_selection,
    make_text_layer,
    paste_image,
    paste_layer,
    remove_layers,
    remove_z,
    select_ellipse,
    select_item,
    select_polygon,
    select_rect,
    shape_image
)
from roller_polygon import (
    CELL_SHAPE_D,
    arrange_hexagon,
    arrange_hexagon_truncated,
    arrange_octagon,
    arrange_octagon_on_its_side,
    calc_hexagon_offset,
    calc_hexagon_truncated_offset,
    calc_octagon_offset,
    calc_octagon_on_its_side_offset,
    calc_parallelogram_left,
    calc_parallelogram_right
)
from roller_one_the import The
import gimpfu as fu
import math

RATIO = sh.TRIANGLE_SCALE_RATIO_DOWN
UP_RATIO = sh.TRIANGLE_SCALE_RATIO_UP
pdb = fu.pdb


def calc_square_rect(o):
    """
    Calculate the dimension and position of a square scaled selection.

    o: One
    Return: tuple
        (w, x, y)
        Define a scaled square.
    """
    w = min(o.scale.w, o.scale.h)
    w1 = w / 2.
    x, y = o.sel.center()
    return w, x - w1, y - w1


def do_circle(j, o):
    """
    Make a circle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    w, x, y = calc_square_rect(o)
    select_ellipse(j, x, y, w, w)


def do_cut_corner_mask(j, o):
    """
    Make a cut corner selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    x, y, w, h = o.sel.rect

    # the difference between scale and selection, 'w1, h1'
    w1 = (w - o.scale.w) / 2.
    h1 = (h - o.scale.h) / 2.

    # eight vertices defining segments
    q = (
        x + w1, y,
        x + w - w1, y,
        x + w, y + h1,
        x + w, y + h - h1,
        x + w - w1, y + h,
        x + w1, y + h,
        x, y + h - h1,
        x, y + h1
    )
    select_polygon(j, q)


def do_darks(j, o):
    """
    Make a selection based on the dark profile of a layer.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    make_polar_mask(j, o, is_dark=True)


def do_ellipse(j, o):
    """
    Make an oval selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_ellipse(j, *o.scale.rect)


def do_ellipse_notch(j, o):
    """
    Cut an ellipse from each corner.

    j: GIMP image
    o: One
    Return: state of selection
    """
    x, y, w, h = o.sel.rect
    scale_w, scale_h = o.scale.size

    # the difference between scale and selection, 'w1, h1'
    w1 = (w - scale_w) / 2.
    h1 = (h - scale_h) / 2.

    # x
    x1 = x - w1
    x2 = x + w - w1

    # y
    y1 = y - h1
    y2 = y + h - h1

    select_rect(j, *o.sel.rect)

    w2 = w1 * 2
    h2 = h1 * 2
    for x, y in ((x1, y1), (x1, y2), (x2, y1), (x2, y2)):
        select_ellipse(j, x, y, w2, h2, option=fu.CHANNEL_OP_SUBTRACT)


def do_eye(j, o):
    """
    Make an eye selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    do_eye_opening(j, o)

    x, y, x1, y1 = get_select_coord(j)
    h = y1 - y
    h1 = h * o.d[ok.PUPIL_SCALE]
    w = x1 - x
    x2 = x + (w - h) / 2.
    offset = (h - h1) / 2.

    select_ellipse(
        j, x2, y + 1., h, max(1., h - 2.), option=fu.CHANNEL_OP_SUBTRACT
    )
    select_ellipse(
        j, x2 + offset, y + offset, h1, h1, option=fu.CHANNEL_OP_ADD
    )


def do_eye_opening(j, o):
    """
    Make an eye selection.

    Ellipse Radius Formula Reference
    rechneronline.de/pi/circular-segment.php

    Formula
    radius = (w² / 4 + h²) / 2 / h

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    w, h = o.scale.w, o.scale.h / 2.
    center_x, center_y = o.sel.center()

    # topleft of eye rectangle, 'x, y'
    x = center_x - w / 2.
    y = center_y - h

    radius = (w * w / 4. + h * h) / 2. / h
    diameter = radius * 2.
    x = x - (diameter - w) / 2.

    select_ellipse(j, x, y, diameter, diameter)
    select_ellipse(
        j,
        x, y - diameter + o.scale.h,
        diameter, diameter,
        option=fu.CHANNEL_OP_INTERSECT
    )


def do_hexagon(j, o):
    """
    Make a hexagon selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    w, h = o.scale.w, o.scale.h
    center_x, center_y = o.sel.center()

    # There are two possible solutions.
    # first solution
    w1, h1 = w, w * UP_RATIO

    # Check for overflow.
    if h1 > h:
        # second solution
        w1, h1 = h * RATIO, h

    q = calc_hexagon_offset(w1, h1)
    select_polygon(
        j,
        arrange_hexagon(q, center_x - q[0], center_y - q[2], w1, h1)
    )


def do_hexagon_shear(j, o):
    """
    Make a sheared hexagon selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(
        j,
        arrange_hexagon(calc_hexagon_offset(*o.scale.size), *o.scale.rect)
    )


def do_hexagon_truncated_shear(j, o):
    """
    Make a sheared truncated hexagon selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(
        j,
        arrange_hexagon_truncated(
            calc_hexagon_truncated_offset(*o.scale.size), *o.scale.rect
        )
    )


def do_hexagon_truncated(j, o):
    """
    Make a truncated hexagon selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    center_x, center_y = o.sel.center()
    w, h = o.scale.w, o.scale.h

    # There are two possible solutions.
    # first solution
    w1, h1 = h * UP_RATIO, h

    # Check for overflow.
    if w1 > w:
        # second solution
        w1, h1 = w, w * RATIO

    q = calc_hexagon_truncated_offset(w1, h1)
    x = center_x - q[1]
    y = center_y - q[3]
    select_polygon(j, arrange_hexagon_truncated(q, x, y, w1, h1))


def do_image(j, o):
    """
    Make an image selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    pdb.gimp_selection_none(j)

    j1 = o.maya.get_mask_image(o.maya.k)
    if j1:
        copy_all_image(j1.j)

        j2 = paste_image()

        # Shape, copy, and delete the image.
        shape_image(j2, *o.scale.size)

        z1 = paste_layer(j.layers[0])

        pdb.gimp_layer_set_offsets(z1, int(o.scale.x), int(o.scale.y))
        select_item(z1)
        remove_z(z1)


def do_lights(j, o):
    """
    Make a selection based on the lights profile of a layer.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    make_polar_mask(j, o)


def do_octagon(j, o):
    """
    Make an octagon selection.

    Reference
        'math.stackexchange.com/
        questions/22064/
        calculating-a-point-that-lies-on-an-ellipse-given-an-angle'

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    w, x, y = calc_square_rect(o)
    q = calc_octagon_offset(w, w)
    select_polygon(j, arrange_octagon(q, x, y, w, w))


def do_octagon_shear(j, o):
    """
    Make a sheared octagon selection.

    Reference to a Helpful Formula
        'math.stackexchange.com/
        questions/22064/
        calculating-a-point-that-lies-on-an-ellipse-given-an-angle'

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(
        j,
        arrange_octagon(calc_octagon_offset(*o.scale.size), *o.scale.rect)
    )


def do_octagon_on_its_side(j, o):
    """
    Make an Octagon on its side selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    w, x, y = calc_square_rect(o)
    q = calc_octagon_on_its_side_offset(w, w)
    select_polygon(j, arrange_octagon_on_its_side(q, x, y, w, w))


def do_octagon_on_its_side_shear(j, o):
    """
    Make a sheared octagon that its on its side selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(
        j,
        arrange_octagon_on_its_side(
            calc_octagon_on_its_side_offset(*o.scale.size), *o.scale.rect
        )
    )


def do_parallelogram_left(j, o):
    """
    Make an parallelogram left selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(
        j, calc_parallelogram_left(o.scale, o.d[ok.PARALLELOGRAM_SCALE])
    )


def do_parallelogram_right(j, o):
    """
    Make an parallelogram right selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(
        j, calc_parallelogram_right(o.scale, o.d[ok.PARALLELOGRAM_SCALE])
    )


def do_rectangle_notch(j, o):
    """
    Make a plus sign selection.

    j: GIMP image
    o: One
    Return: state of selection
    """
    x, y, w, h = o.sel.rect
    scale_w, scale_h = o.scale.size

    # the difference between scale and selection, 'w1, h1'
    w1 = (w - scale_w) / 2.
    h1 = (h - scale_h) / 2.
    x1 = x + w1
    x2 = x1 + scale_w
    x3 = x + w
    y1 = y + h1
    y2 = y1 + scale_h
    y3 = y + h
    select_polygon(j, (
        x, y1,
        x1, y1,
        x1, y,
        x2, y,
        x2, y1,
        x3, y1,
        x3, y2,
        x2, y2,
        x2, y3,
        x1, y3,
        x1, y2,
        x, y2
    ))


def do_rectangle(j, o):
    """
    Make a rectangle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_rect(j, *o.scale.rect)


def do_square_mitered(j, o):
    """
    Make a square, rotated by 45 degrees, selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    w = min(o.scale.w, o.scale.h) / 2.
    x, y = o.sel.center()
    q = (
        x, y - w,
        x + w, y,
        x, y + w,
        x - w, y
    )
    select_polygon(j, q)


def do_diamond(j, o):
    """
    Make a sheared diamond selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    w, h = o.scale.w / 2., o.scale.h / 2.
    x, y = o.sel.center()

    # four vertices defining segments
    q = (
        x, y - h,
        x + w, y,
        x, y + h,
        x - w, y
    )
    select_polygon(j, q)


def do_rounded_corner(j, o):
    """
    Make a rounded Corner selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    x, y = o.sel.x, o.sel.y
    w = o.sel.w - o.scale.w
    h = o.sel.h - o.scale.h

    # topleft
    select_ellipse(j, x, y, w, h)

    # top-right
    x1 = x + o.sel.w - w

    select_ellipse(j, x1, y, w, h, option=fu.CHANNEL_OP_ADD)

    # bottom-left
    y1 = y + o.sel.h - h

    select_ellipse(j, x, y1, w, h, option=fu.CHANNEL_OP_ADD)

    # bottom-right
    select_ellipse(j, x1, y1, w, h, option=fu.CHANNEL_OP_ADD)

    # vertical rectangle
    x2 = x + w / 2.
    w1 = o.sel.w - w

    select_rect(j, x2, y, w1, o.sel.h, option=fu.CHANNEL_OP_ADD)

    # horizontal rectangle
    y2 = y + h / 2.
    h1 = o.sel.h - h
    select_rect(j, x, y2, o.sel.w, h1, option=fu.CHANNEL_OP_ADD)


def do_square(j, o):
    """
    Make a square selection.

    j: GIMP image
        to receive selection

    v: View
    maya: Maya
    Return: state of selection
    """
    w, x, y = calc_square_rect(o)
    select_rect(j, x, y, w, w)


def do_text(j, o):
    """
    Make a text selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    cat = The.cat
    font = o.d[ok.FONT]

    if font not in cat.font_list:
        font = cat.font_list[0] if cat.font_list else None
    if font is not None:
        font_size = max(o.scale.w, o.scale.h)
        character = o.d[ok.TEXT]
        if character:
            z = add_layer(j, "text")
            go, z1 = make_text_layer(
                j, z,
                0,                      # no antialias
                character,
                font_size,
                font,
                (255, 255, 255),
                o.sel.x, o.sel.y,
                o.scale.w, o.scale.h
            )
            if go:
                select_item(z1)

                is_sel, x, y, x1, y1 = get_select_bounds(j)

                if is_sel:
                    center_x, center_y = o.sel.center()
                    x = center_x - (x1 - x) / 2.
                    y = center_y - (y1 - y) / 2.

                    pdb.gimp_layer_set_offsets(z1, int(x), int(y))
                    select_item(z1)
                remove_layers((z, z1))


def do_triangle_bottom_left(j, o):
    """
    Select half of a scaled rectangle with a triangle
    positioned in the bottom-left corner of the rectangle.

    j: GIMP image
    o: One
    Return: state of selection
    """
    x, y, w, h = o.scale.rect
    x1 = x + w
    y1 = y + h
    select_polygon(j, (x, y, x1, y1, x, y1))


def do_triangle_bottom_right(j, o):
    """
    Select half of a scaled rectangle with a triangle
    positioned in the bottom-right corner of the rectangle.

    j: GIMP image
    o: One
    Return: state of selection
    """
    x, y, w, h = o.scale.rect
    x1 = x + w
    y1 = y + h
    select_polygon(j, (x1, y, x1, y1, x, y1))


def do_triangle_topleft(j, o):
    """
    Select half of a scaled rectangle with a triangle
    positioned in the topleft corner of the rectangle.

    j: GIMP image
    o: One
    Return: state of selection
    """
    x, y, w, h = o.scale.rect
    x1 = x + w
    y1 = y + h
    select_polygon(j, (x, y, x1, y, x, y1))


def do_triangle_top_right(j, o):
    """
    Select half of a scaled rectangle with a triangle
    positioned in the top-right corner of the rectangle.

    j: GIMP image
    o: One
    Return: state of selection
    """
    x, y, w, h = o.scale.rect
    x1 = x + w
    y1 = y + h
    select_polygon(j, (x, y, x1, y, x1, y1))


def do_triangle_down_regular(j, o):
    """
    Make a down-facing regular triangle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(j, CELL_SHAPE_D[ft.TRIANGLE_DOWN_REGULAR](o.scale))


def do_triangle_down_shear(j, o):
    """
    Make a sheared down-facing triangle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(j, CELL_SHAPE_D[ft.TRIANGLE_DOWN_SHEAR](o.scale))


def do_triangle_left_regular(j, o):
    """
    Make a left-facing regular triangle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(j, CELL_SHAPE_D[ft.TRIANGLE_LEFT_REGULAR](o.scale))


def do_triangle_left_shear(j, o):
    """
    Make a sheared left-facing triangle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(j, CELL_SHAPE_D[ft.TRIANGLE_LEFT_SHEAR](o.scale))


def do_triangle_right_regular(j, o):
    """
    Make a right-facing regular triangle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(j, CELL_SHAPE_D[ft.TRIANGLE_RIGHT_REGULAR](o.scale))


def do_triangle_right_shear(j, o):
    """
    Make a sheared right-facing triangle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(j, CELL_SHAPE_D[ft.TRIANGLE_RIGHT_SHEAR](o.scale))


def do_triangle_up_regular(j, o):
    """
    Make an up-facing regular triangle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(
        j, CELL_SHAPE_D[ft.TRIANGLE_UP_REGULAR](o.scale))


def do_triangle_up_shear(j, o):
    """
    Make a sheared up-facing triangle selection.

    j: GIMP image
        to receive selection

    o: One
    Return: state of selection
    """
    select_polygon(j, CELL_SHAPE_D[ft.TRIANGLE_UP_SHEAR](o.scale))


def dispatch_corner(j, o):
    """
    Dispatch corner mask function depending on the Corner Type.

    j: Image
    o: One
    Return: state of selection
    """
    p = CORNER_D.get(o.d[ok.CORNER_TYPE])
    if p:
        p(j, o)


def dispatch_hexagon(j, o):
    """
    Dispatch Hexagon Mask function.

    j: Image
    o: One
    Return: state of selection
    """
    p = HEXAGON_D.get(o.d[ok.HEXAGON_TYPE])
    if p:
        p(j, o)


def dispatch_octagon(j, o):
    """
    Dispatch Octagon Mask function.

    j: Image
    o: One
    Return: state of selection
    """
    p = OCTAGON_D.get(o.d[ok.OCTAGON_TYPE])
    if p:
        p(j, o)


def dispatch_rectangle(j, o):
    """
    Dispatch Rectangle Mask function.

    j: Image
    o: One
    Return: state of selection
    """
    p = RECTANGLE_D.get(o.d[ok.RECTANGLE_TYPE])
    if p:
        p(j, o)


def dispatch_triangle(j, o):
    """
    Dispatch Triangle Mask function.

    j: Image
    o: One
    Return: state of selection
    """
    p = TRIANGLE_D.get(o.d[ok.TRIANGLE_TYPE])
    if p:
        p(j, o)


def feather_mask(j, d):
    """
    Feather a Mask selection.

    j: GIMP image
        Is the render.

    d: dict
        Mask Preset
        {Option key: value}

    Return: state of selection
    """
    if not pdb.gimp_selection_is_empty(j) and d[ok.FEATHER]:
        sel = pdb.gimp_selection_save(j)
        a = d[ok.FEATHER]
        f = a / d[ok.STEPS]
        b = f
        if a:
            while a > b:
                pdb.gimp_selection_feather(j, b)
                b = min(b + f, a)

        load_selection(j, sel, option=fu.CHANNEL_OP_INTERSECT)
        pdb.gimp_image_remove_channel(j, sel)


def make_polar_mask(j, o, is_dark=False):
    """
    Make a selection based on a lights or darks profile of a layer.

    j: GIMP image
        to receive selection

    o: One
    is_dark: bool
        If it is True, the mask hides the light-side pixels.
    """
    z = o.maya.matter
    mask = pdb.gimp_layer_create_mask(z, fu.ADD_MASK_COPY)

    pdb.gimp_layer_add_mask(z, mask)

    if is_dark:
        # no linear, '0'
        pdb.gimp_drawable_invert(mask, 0)

    do_rectangle(j, o)
    pdb.gimp_image_select_item(j, fu.CHANNEL_OP_INTERSECT, mask)
    discard_mask(z)


def rotate_mask(j, d):
    """
    Rotate a Mask selection.

    j: GIMP image
        Is the render.

    d: dict
        Mask Preset

    Return: state of selection
    """
    if not pdb.gimp_selection_is_empty(j) and d[ok.ANGLE]:
        z = add_layer(j, "Rotate")

        color_fill_selection(z, (0, 0, 0))

        # auto-center, 'True'; x and y, '0'
        z1 = pdb.gimp_item_transform_rotate(
            z, math.radians(d[ok.ANGLE]), True, .0, .0
        )

        select_item(z1)
        remove_layers((z, z1))


CORNER_D = {
    ms.CUT_CORNER: do_cut_corner_mask,
    ms.ELLIPSE_NOTCH: do_ellipse_notch,
    ms.RECTANGLE_NOTCH: do_rectangle_notch,
    ms.ROUNDED_CORNER: do_rounded_corner
}
HEXAGON_D = {
    sh.HEXAGON: do_hexagon,
    sh.HEXAGON_SHEAR: do_hexagon_shear,
    sh.HEXAGON_TRUNCATED: do_hexagon_truncated,
    sh.HEXAGON_TRUNCATED_SHEAR: do_hexagon_truncated_shear
}
OCTAGON_D = {
    sh.OCTAGON: do_octagon,
    sh.OCTAGON_SHEAR: do_octagon_shear,
    sh.OCTAGON_ON_ITS_SIDE: do_octagon_on_its_side,
    sh.OCTAGON_ON_ITS_SIDE_SHEAR: do_octagon_on_its_side_shear,
}
RECTANGLE_D = {
    sh.RECTANGLE: do_rectangle,
    sh.SQUARE: do_square,
    sh.SQUARE_MITERED: do_square_mitered
}
TRIANGLE_D = {
    ft.TRIANGLE_DOWN_REGULAR: do_triangle_down_regular,
    ft.TRIANGLE_DOWN_SHEAR: do_triangle_down_shear,
    ft.TRIANGLE_LEFT_REGULAR: do_triangle_left_regular,
    ft.TRIANGLE_LEFT_SHEAR: do_triangle_left_shear,
    ft.TRIANGLE_RIGHT_REGULAR: do_triangle_right_regular,
    ft.TRIANGLE_RIGHT_SHEAR: do_triangle_right_shear,
    ft.TRIANGLE_UP_REGULAR: do_triangle_up_regular,
    ft.TRIANGLE_UP_SHEAR: do_triangle_up_shear,
    ms.TRIANGLE_BOTTOM_LEFT: do_triangle_bottom_left,
    ms.TRIANGLE_BOTTOM_RIGHT: do_triangle_bottom_right,
    ms.TRIANGLE_TOPLEFT: do_triangle_topleft,
    ms.TRIANGLE_TOP_RIGHT: do_triangle_top_right
}
MASK_FUNCTION = {
    sh.CIRCLE: do_circle,
    ms.CORNER: dispatch_corner,
    ms.DARKS: do_darks,
    ms.ELLIPSE: do_ellipse,
    ms.EYE: do_eye,
    ms.EYE_OPENING: do_eye_opening,
    ms.IMAGE: do_image,
    ms.LIGHTS: do_lights,
    ms.PARALLELOGRAM_LEFT: do_parallelogram_left,
    ms.PARALLELOGRAM_RIGHT: do_parallelogram_right,
    ms.TEXT: do_text,
    ms.TRIANGLE: dispatch_triangle,
    sh.DIAMOND: do_diamond,
    sh.HEXAGON: dispatch_hexagon,
    sh.OCTAGON: dispatch_octagon,
    sh.RECTANGLE: dispatch_rectangle
}
