#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_for import Frame as ff
from roller_constant_key import Option as ok
from roller_frame import Noise, Overlay, grow_frame
from roller_frame_build import Build
from roller_fu import (
    add_layer,
    clone_layer,
    color_fill_layer,
    color_fill_selection,
    load_selection,
    select_item
)
from roller_fu_mode import get_mode
from roller_maya import check_matter, check_mix_wrap, make_frame_group
from roller_maya_blur_behind import BlurBehind
from roller_maya_light import Light
from roller_maya_shadow import Shadow
from roller_view_hub import PROFILE
from roller_view_real import (
    LIGHT, OVERLAY, add_wip_layer, get_light, get_noise, get_overlay
)
import gimpfu as fu

pdb = fu.pdb


def adjust_opacity(z):
    """
    For a given layer's RGB pixel, adjust its alpha channel
    mean intensity value to be above a threshold (225).

    z: layer
        with material to adjust

    Return: layer, Selection state
        adjusted
    """
    j = z.image

    # the mean intensity value, 'f'
    f = 0

    while f < 225.:
        # Create a selection-mask.
        select_item(z)

        if not pdb.gimp_selection_is_empty(j):
            # Get the mean intensity value of the material's alpha, 'f'.
            f = pdb.gimp_drawable_histogram(z, fu.HISTOGRAM_ALPHA, .0, 1.)[0]
            if f < 225.:
                # double-down to double the alpha value
                z = clone_layer(z)
                z = pdb.gimp_image_merge_down(j, z, fu.CLIP_TO_IMAGE)
        else:
            break
    return z


def do_color(v, maya):
    """
    Make a color layer.

    v: View
    maya: Maya
    Return: layer
        with color
    """
    pdb.gimp_selection_none(v.j)

    d = maya.value_d
    a = maya.super_maya
    z = add_wip_layer(v, "Color", a.group, offset=get_light(a) + get_noise(a))

    color_fill_layer(z, d[ok.COLOR_1])
    return z


def do_matter(v, maya):
    """
    Make a frame.

    v: View
    maya: Maya
    Return: layer or None
        with a frame
    """
    j = v.j
    d = maya.value_d[ok.WRW][ok.WRAP_CP]
    group = maya.group
    z = add_layer(
        j,
        group.name + " Color Pipe",
        parent=group,
        offset=get_light(maya) + get_noise(maya) + get_overlay(maya)
    )
    w = d[ok.WIDTH]
    q = PROFILE[d[ok.PROFILE]](w)

    select_item(maya.cause.matter)
    draw_profile(z, w, q)

    z = adjust_opacity(z)
    z.opacity = d[ok.OPACITY]
    z.mode = get_mode(d, k=ok.MODE)
    return z


def draw_profile(z, w, q):
    """
    Draw a color-type gradient frame using a profile.

    z: layer
        for blur material

    w: float
        width of border

    q: list
        of profile
        in .0 to 1.
    """
    j = z.image

    # selections
    a = b = None

    for i in range(int(w)):
        if a:
            # Do now to save memory for large frames.
            pdb.gimp_image_remove_channel(j, a)

        a = b = pdb.gimp_selection_save(j)
        if b:
            grow_frame(j, 1, ff.ANGULAR)
            load_selection(j, b, option=fu.CHANNEL_OP_SUBTRACT)
            color = int(round(q[i] * 255.))
            color_fill_selection(z, (color, color, color))
            load_selection(j, b, option=fu.CHANNEL_OP_ADD)
    if b:
        pdb.gimp_image_remove_channel(j, b)


class ColorPipe(Build):
    """Create a frame with a custom profile and color."""
    issue_q = 'matter', 'mode', 'opacity', 'shade'
    put = (
        (make_frame_group, 'group'),
        (check_matter, 'matter'),
        (check_mix_wrap, None)
    )
    wrap_k = ok.WRAP_CP

    def __init__(self, any_group, super_maya, k_path):
        """
        any_group: AnyGroup
            Owns the option.

        super_maya: Maya
            super type
            Is from the AnyGroup where the Frame Widget resides.

        k_path: tuple
            Key path of the Frame Button in its vote dict.
            (Option key, ...)
        """
        Build.__init__(
            self, any_group, super_maya, k_path + (ok.WRW, ok.WRAP_CP), do_matter
        )

        self.sub_maya[OVERLAY] = Overlay(
            any_group, self, k_path + (ok.WRW, ok.OVERLAY_CF), do_color
        )
        self.sub_maya[ok.NOISE_D] = Noise(
            any_group, self, k_path + (ok.SRW, ok.NOISE_D)
        )
        self.sub_maya[ok.SHADOW] = Shadow(
            any_group, self, k_path + (ok.SRW, ok.SHADOW), (self.cause, self)
        )
        self.sub_maya[ok.BLUR_BEHIND] = BlurBehind(
            any_group, self, k_path, ok.SRW
        )
        self.sub_maya[LIGHT] = Light(any_group, self, ok.TRANSLUCENT)

    def do(self, v, d, is_change):
        """
        Check, modify, and produce layer output.

        v: View
        d: dict
            Color Pipe Preset
            {Option key: value}

        is_change: bool
            Is the state of the super Maya's matter and/or mask.

        Return: bool
            If True, then Shadow changed the background.
        """
        self.value_d = d
        is_back = v.is_back
        self.is_matter |= is_change

        self.realize(v)
        self.sub_maya[OVERLAY].do(
            v, d[ok.WRW][ok.OVERLAY_CF], is_change, self.is_matter
        )
        self.sub_maya[ok.NOISE_D].do(
            v, d[ok.SRW][ok.NOISE_D], is_change, self.is_matter
        )

        m = self.sub_maya[ok.SHADOW].do(
            v, d[ok.SRW][ok.SHADOW], self.is_matter, is_change
        )

        self.sub_maya[ok.BLUR_BEHIND].do(
            v, d[ok.SRW][ok.BLUR_BEHIND], m or is_back, self.is_matter
        )
        self.sub_maya[LIGHT].do(v, self.is_matter)
        self.reset_issue()
        return m
