#!/usr/bin/env python
# -*- coding: utf-8 -*-
from math import atan2, cos, sin, radians
from random import randint, uniform
from roller_constant_key import Option as ok
from roller_frame import Overlay, ShadowBasic
from roller_frame_build import Build
from roller_maya import check_matter, make_frame_group
from roller_maya_blur_behind import BlurBehind
from roller_maya_light import Light
from roller_fu import (
    add_layer,
    blur_selection,
    clear_inverse_selection,
    clear_selection,
    clone_layer,
    color_fill_selection,
    color_fill_layer,
    get_item_size,
    get_select_coord,
    invert_selection,
    load_selection,
    make_layer_group,
    merge_layer,
    merge_layer_group,
    select_item,
    select_polygon,
    verify_layer
)
from roller_one_gegl import noise_rgb
from roller_one_rect import Rect
from roller_view_preset import combine_seed
from roller_view_real import (
    OVERLAY, LIGHT, SHADOW, add_wip_layer, get_overlay, get_light
)
import gimpfu as fu

MAIN = 'main'
pdb = fu.pdb


def calc_angle_shift(d):
    """
    Randomize the angle-shift amount.

    d: dict
        Corner Tape Preset
        {Option key: value}

    Return: float
        angle shift
    """
    f = d[ok.ANGLE_SHIFT]
    return uniform(radians(-f), radians(f))


def calc_rotated_bounds(maya, tape):
    """
    Calculate the rotated bounds of the
    work-in-progress image in its original size.

    Calculate the transform amount for the x, y vectors.
    The goal of the transform is to calculate
    the corner points of the a placed image.

    The function only works when the original image size ratio
    is the same as the cell image's size ratio. The Locked Resize
    Method will leave the ratio the same, but the other Resize
    Methods typically modify this ratio.

    maya: CornerTape
    tape: Tape
    """
    q = tape.corners
    j = maya.cause.get_image(tape.k)

    if j:
        w, h = get_item_size(j)
        f = tape.angle

        # center
        w, h = w / 2., h / 2.

        radius = ((w**2.) + (h**2.))**.5

        for i in range(4):
            if not i:
                x, y = -w, h

            elif i == 1:
                x, y = w, h

            elif i == 2:
                x, y = w, -h

            else:
                x, y = -w, -h

            x, y = calc_point(.0, .0, atan2(x, y) + f, radius)
            q.append((x, y))

        q_x = [i[0] for i in q]
        q_y = [i[1] for i in q]
        u = min(q_x), min(q_y)
        u1 = max(q_x), max(q_y)
        u = abs(u[0] - u1[0]), abs(u[1] - u1[1])
        tape.transform = tape.rect.w / u[0], tape.rect.h / u[1]


def calc_point(x, y, angle, radius):
    """
    Return a point on a circle that intersects with at a
    vector starting from the center of the circle at a angle.

    x, y: float
        center point

    angle : float
        rotation angle in radians

    radius: float
        the radius of the circle in pixels

    Return: tuple
        (x, y) of float
        the point on the circle
    """
    x = (sin(angle) * radius) + x
    y = (cos(angle) * -radius) + y
    return round(x), round(y)


def calc_tape_length(d):
    """
    Calculate tape length.

    d: dict
        Has options.

    Return: float
        tape length
    """
    return max(
        1.,
        d[ok.LENGTH] + randint(-d[ok.LENGTH_SHIFT], d[ok.LENGTH_SHIFT])
    )


def do_color(v, maya):
    """
    Make a color overlay layer.

    v: View
    maya: CornerTape
    Return: layer or None
        with material
    """
    d = maya.value_d
    a = maya.super_maya
    z = add_wip_layer(v, "Overlay", a.group, offset=get_light(a))

    color_fill_layer(z, d[ok.COLOR_1])
    return z


def do_matter(v, maya):
    """
    Make a tape layer.

    v: View
    maya: CornerTape
    Return: layer or None
        with material
    """
    j = v.j
    z = None
    parent = maya.group
    offset = get_overlay(maya) + get_light(maya)
    super_ = maya.cause

    if super_.vote_type == MAIN:
        group = make_layer_group(j, "Group", parent=parent, offset=offset)
        model = maya.model

        for k in super_.main_q:
            sel = model.get_image_sel(k)

            load_selection(j, sel)
            if not pdb.gimp_selection_is_empty(j):
                do_sel(
                    v,
                    maya,
                    add_layer(
                        j, "Base", parent=group, offset=len(group.layers)
                    ),
                    k
                )
        z = merge_layer_group(group)

    else:
        select_item(maya.cause.matter)
        if not pdb.gimp_selection_is_empty(j):
            z = add_layer(j, "Base", parent=parent, offset=offset)
            z = do_sel(v, maya, z, maya.cause.k)

    z = verify_layer(z)

    if z:
        z.name = maya.group.name + " Corner Tape"
        z.opacity = 34.
        z.mode = fu.LAYER_MODE_DARKEN_ONLY
    return z


def do_sel(v, maya, z, k):
    """
    Draw tape. Select the image material prior to calling.

    v: View
    maya: CornerTape
    k: tuple or None
        (row, column)

    Return: layer or None
        with material
    """
    j = v.j
    a = Tape(maya, k, *get_select_coord(j))

    if a.angle:
        calc_rotated_bounds(maya, a)

    b = a.rect

    # noise layer, 'z1'
    z1 = clone_layer(z)
    z1.opacity = 50.

    combine_seed(v, a.d)
    pdb.gimp_selection_none(j)
    noise_rgb(
        z1,
        uniform(.001, .04),
        uniform(.001, .04),
        uniform(.001, .04),
        1.,                             # alpha
        int(a.d[ok.SEED] + v.glow_ball.seed)
    )
    select_topleft(v, a, b.x, b.y)
    select_top_right(v, a, b.x + b.w, b.y)
    select_bottom_left(v, a, b.x, b.y + b.h)
    select_bottom_right(v, a, b.x + b.w, b.y + b.h)
    color_fill_selection(z, (255, 255, 255))
    clear_inverse_selection(z1)

    for i in range(2):
        pdb.plug_in_shift(j, z, 1, 0)
        pdb.plug_in_shift(j, z, 1, 1)

    blur_selection(z, 2.)

    z = merge_layer(z1)

    # Darken the edge.
    select_item(z)
    pdb.gimp_selection_shrink(j, 1)
    invert_selection(j)
    pdb.gimp_drawable_curves_spline(
        z,
        fu.HISTOGRAM_VALUE,
        4,                              # coordinate count
        (.0, .0392, 1., .3137)
    )
    return z


def get_corner_point(tape, x, y, corner_x):
    """
    tape: Tape
    x, y: int
        corner point

    corner_x: int
        index to corner point

    Return: tuple
        of int
        corner point
    """
    def _shift():
        """
        Randomize corner-shift amount.

        Return: int
            corner shift
        """
        _a = a.d[ok.CORNER_SHIFT]
        return randint(-_a, _a)

    a = tape

    if a.angle:
        x1, y1 = a.corners[corner_x]
        x1 *= a.transform[0]
        y1 *= a.transform[1]
        x = x1 + a.center[0]
        y = y1 + a.center[1]

    x += _shift()
    y += _shift()
    return x, y


def select_tape_rect(v, d, angle, x, y, w):
    """
    Define and select a tape rectangle.

    v: View
    d: dict
        Corner Tape Preset
        {Option key: value}

    x, y: int
        start point

    w: int
        length of tape
    """
    w1 = d[ok.WIDTH]

    # 90 degrees in radianm, 1.5708
    x1, y1 = calc_point(x, y, angle - 1.5708, w1)

    # 180 degrees in radian, 3.14159
    x2, y2 = calc_point(x1, y1, angle - 3.14159, w)

    # 270 degrees in radian, 4.71239
    x3, y3 = calc_point(x2, y2, angle - 4.71239, w1)

    select_polygon(
        v.j,
        (x, y, x1, y1, x2, y2, x3, y3),
        option=fu.CHANNEL_OP_ADD
    )


def select_bottom_left(v, tape, x, y):
    """
    Make a selection for the bottom-left corner.

    v: View
    tape: Tape
        Has scope variable.

    x, y: float
        corner point
    """
    d = tape.d

    # 135 degrees in radian, 2.35619
    angle = 2.35619 + calc_angle_shift(d) + tape.angle

    x, y = get_corner_point(tape, x, y, 3)
    w = calc_tape_length(d)
    x1, y1 = calc_point(x, y, angle, w / 2.)
    select_tape_rect(v, d, angle, x1, y1, w)


def select_bottom_right(v, tape, x, y):
    """
    Make a selection for the bottom-right corner.

    v: View
    tape: Tape
    x, y: int
        corner point
    """
    d = tape.d

    # 45 degees in radian, .785398
    angle = .785398 + calc_angle_shift(d) + tape.angle

    x, y = get_corner_point(tape, x, y, 2)
    w = calc_tape_length(d)
    x1, y1 = calc_point(x, y, angle, w / 2.)
    select_tape_rect(v, d, angle, x1, y1, w)


def select_topleft(v, tape, x, y):
    """
    Make a selection for the topleft corner.

    v: View
    tape: Tape
    x, y: int
        corner point
    """
    d = tape.d

    # 225 degrees in radian, 3.92699
    angle = 3.92699 + calc_angle_shift(d) + tape.angle

    x, y = get_corner_point(tape, x, y, 0)
    w = calc_tape_length(d)
    x1, y1 = calc_point(x, y, angle, w / 2.)
    select_tape_rect(v, d, angle, x1, y1, w)


def select_top_right(v, tape, x, y):
    """
    Make a selection for the top-right corner.

    v: View
    tape: Tape
    x, y: float
        center point
    """
    d = tape.d

    # 315 degrees in radian, 5.49779
    angle = 5.49779 + calc_angle_shift(d) + tape.angle

    x, y = get_corner_point(tape, x, y, 1)
    w = calc_tape_length(d)
    x1, y1 = calc_point(x, y, angle, w / 2.)
    select_tape_rect(v, d, angle, x1, y1, w)


class CornerTape(Build):
    """Simulate a tape holding an image in place."""
    is_seeded = True
    issue_q = 'matter', 'mode', 'opacity'
    put = (make_frame_group, 'group'), (check_matter, 'matter')

    def __init__(self, any_group, super_maya, k_path):
        """
        any_group: AnyGroup
            owner

        super_maya: Maya
            enclosing type
            Is from the AnyGroup where the Frame Widget resides.

        k_path: tuple
            Key path of the Frame Button in its vote dict.
            (Option key, ...)
        """
        Build.__init__(
            self, any_group, super_maya, k_path + (ok.WRW, ok.TAPE), do_matter
        )

        self.sub_maya[OVERLAY] = Overlay(
            any_group, self, k_path + (ok.WRW, ok.OVERLAY_CF), do_color
        )
        self.sub_maya[SHADOW] = ShadowBasic(
            any_group, self, k_path + (ok.SRW, ok.SHADOW_BASIC)
        )
        self.sub_maya[LIGHT] = Light(any_group, self, ok.TRANSLUCENT)
        self.sub_maya[ok.BLUR_BEHIND] = BlurBehind(
            any_group, self, k_path, ok.SRW
        )

    def do(self, v, d, is_change):
        """
        Produce and modify layer output.

        v: View
        d: dict
            Corner Tape Preset
            {Option key: value}

        is_change: bool
            Is the state of the super Maya's matter and/or mask.

        Return: False
            Shadow did not change the background.
        """
        self.value_d = d
        is_back = v.is_back
        self.is_matter |= is_change

        self.realize(v)
        self.sub_maya[OVERLAY].do(
            v, d[ok.WRW][ok.OVERLAY_CF], is_change, self.is_matter
        )
        m = self.sub_maya[SHADOW].do(
            v, d[ok.SRW][ok.SHADOW_BASIC], self.is_matter
        )

        if m or self.is_matter:
            # Remove shadow from behind tape.
            select_item(self.matter)
            clear_selection(self.sub_maya[SHADOW].shadow)

        self.sub_maya[LIGHT].do(v, self.is_matter)
        self.sub_maya[ok.BLUR_BEHIND].do(
            v, d[ok.SRW][ok.BLUR_BEHIND], m or is_back, self.is_matter
        )
        self.reset_issue()
        return False


class Tape:
    """Is a container for making Corner Tape material."""

    def __init__(self, maya, k, x, y, x1, y1):
        """
        v: View
        maya: Maya
        k: tuple or None
            cell index for WIP image

        x, y: float
            topleft corner of the image selection

        x1, y1: float
            bottom-right corner of the image selection
        """
        # Is a sequence of four x, y coordinate tuple.
        # topleft, top-right,bottom-right, bottom-left
        self.corners = []

        # Image selection rectangle, 'rect'
        w = x1 - x
        h = y1 - y
        self.rect = Rect(x, y, w, h)

        # Image selection center point, 'center'
        self.center = x + w / 2., y + h / 2.

        # image angle in the Image Preset
        self.angle = radians(maya.cause.value_d[ok.ANGLE])

        # float
        # Use when an image has rotation to transform its corner points.
        # Set to failure value, '1.'.
        self.transform = 1., 1.

        # dict
        # Corner Tape Preset
        self.d = maya.value_d[ok.WRW][ok.TAPE]

        # tuple or None
        # cell index
        # Is None for Canvas.
        self.k = k
