#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_key import Option as ok
from roller_frame import ShadowBasic
from roller_frame_build import Build
from roller_maya import check_matter, make_frame_group
from roller_fu import (
    add_layer,
    blur_selection,
    clear_inverse_selection,
    color_fill_selection,
    color_fill_layer,
    discard_mask,
    isolate_selection,
    make_layer_group,
    merge_layer_group,
    select_item
)
from roller_maya_blur_behind import BlurBehind
from roller_maya_light import Light
from roller_one_gegl import emboss
from roller_view_real import (
    LIGHT, SHADOW, clip_to_wip, clone_background, get_light
)
import gimpfu as fu

pdb = fu.pdb


def select_bump_color(z):
    """
    Create a selection based on color.

    z: layer
        to color

    q: tuple
        RGB color
    option: gimpfu enum
        Selection type
    """
    pdb.gimp_context_set_feather(1)
    pdb.gimp_context_set_feather_radius(3., 3.)
    pdb.gimp_context_set_sample_merged(0)

    # composite
    pdb.gimp_context_set_sample_criterion(0)
    pdb.gimp_context_set_sample_threshold(.33)
    pdb.gimp_context_set_sample_transparent(1)
    pdb.gimp_image_select_color(
        z.image, fu.CHANNEL_OP_REPLACE, z, (255, 255, 255)
    )


def select_bg_white(z):
    """
    Create a selection based on color.

    z: layer
        to color

    q: tuple
        RGB color
    option: gimpfu enum
        Selection type
    """
    pdb.gimp_context_set_feather_radius(4., 4.)

    # composite
    pdb.gimp_context_set_sample_threshold(.33)
    pdb.gimp_image_select_color(
        z.image, fu.CHANNEL_OP_REPLACE, z, (255, 255, 255)
    )


def do_matter(v, maya):
    """
    Make a frame.

    v: View
    maya: Hot Glue Maya
    Return: layer
        with the frame
    """
    j = v.j
    d = maya.value_d
    w, h = j.width, j.height
    cause = maya.cause.matter

    # A mask interferes with 'clone_background'.
    discard_mask(maya.group)

    # Resize material layer so that the waves plug-in will perform
    # its function off-center which has less symmetry and stretch.
    pdb.gimp_image_resize(j, w + w, h + h, 0, 0)

    group = make_layer_group(
        j, "Hot Glue WIP", parent=maya.group, offset=get_light(maya)
    )
    bump_z = add_layer(j, "Bump", parent=group)
    w1 = d[ok.WIDTH]
    w1 = int(w1 // 2. + w1 % 2)

    color_fill_layer(bump_z, (48, 48, 48))
    select_item(cause)

    # Make a border overlapping the cause material.
    pdb.gimp_selection_border(j, int(w1))

    color_fill_selection(bump_z, (255, 255, 255))

    bg_z = clone_background(v, bump_z, n="Base")

    pdb.gimp_selection_none(j)
    pdb.plug_in_waves(
        j, bump_z,
        d[ok.WAVE_AMPLITUDE],
        d[ok.WAVE_PHASE],
        d[ok.WAVELENGTH],
        0,                      # smeared
        0                       # not reflective
    )
    pdb.gimp_image_resize(j, w, h, 0, 0)
    blur_selection(bump_z, d[ok.SOFTEN] * 2.)
    select_bump_color(bump_z)

    sel = pdb.gimp_selection_save(j)

    pdb.gimp_selection_none(j)
    emboss(bump_z, v.glow_ball.azimuth, 45., int(d[ok.SOFTEN]))
    isolate_selection(bump_z, sel)
    pdb.gimp_image_remove_channel(j, sel)

    bg_z.mode = fu.LAYER_MODE_SOFTLIGHT

    blur_selection(bg_z, d[ok.SOFTEN] * 2)
    select_bg_white(bg_z)
    select_item(bump_z, fu.CHANNEL_OP_INTERSECT)
    clear_inverse_selection(bg_z)
    pdb.gimp_image_reorder_item(j, bg_z, group, 0)

    group.mode = fu.LAYER_MODE_OVERLAY
    z = merge_layer_group(group)
    z.name = maya.group.name + " Hot Glue"

    clip_to_wip(v, z)
    return z


class HotGlue(Build):
    """Create a translucent type of frame that has a gluey tube appearance."""
    is_embossed = True
    issue_q = 'matter',
    put = (make_frame_group, 'group'), (check_matter, 'matter')

    def __init__(self, any_group, super_maya, k_path):
        """
        any_group: AnyGroup
            Owns the option.

        super_maya: Maya
            super type
            Is from the AnyGroup where the Frame Widget resides.

        k_path: tuple
            Key path of the Frame Button in its Preset vote dict.
            (Option key, ...)
        """
        Build.__init__(self, any_group, super_maya, k_path, do_matter)
        self.sub_maya[ok.BLUR_BEHIND] = BlurBehind(
            any_group, self, k_path, ok.SRW
        )
        self.sub_maya[SHADOW] = ShadowBasic(
            any_group, self, k_path + (ok.SRW, ok.SHADOW_BASIC)
        )
        self.sub_maya[LIGHT] = Light(any_group, self, ok.TRANSLUCENT)

    def do(self, v, d, is_change):
        """
        Check, modify, and produce layer output.

        v: View
        d: dict
            Hot Glue Preset
            {Option key: value}

        is_change: bool
            Is the state of the super Maya's matter and/or mask.

        Return: False
            Shadow did not change the background.
        """
        self.value_d = d
        is_back = v.is_back
        self.is_matter |= is_change or v.is_back

        self.realize(v)

        m = self.sub_maya[SHADOW].do(
            v, d[ok.SRW][ok.SHADOW_BASIC], self.cause.is_matter
        )

        self.sub_maya[LIGHT].do(v, self.is_matter)
        self.sub_maya[ok.BLUR_BEHIND].do(
            v, d[ok.SRW][ok.BLUR_BEHIND], m or is_back, self.is_matter
        )
        self.reset_issue()
        return False
