#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_for import (
    Deco as dc, Frame as ff, Issue as vo, Signal as si
)
from roller_constant_key import Frame as ek, Option as ok
from roller_fu import (
    make_layer_group, remove_z, set_layer_attr, set_layer_mode
)
from roller_fu_mode import translate_mode
from roller_model import get_main_q
from roller_one_extract import get_option_list_choice
from roller_one_image import get_image_ref
from roller_one_the import The
from roller_view_contain import Deco
from roller_view_real import remove_maya_z
import gimpfu as fu
import os

MAIN = 'main'
NO_VOTE = 'NO VOTE'
PER = 'per'
pdb = fu.pdb


def assign_deco_image(v, d):
    """
    For every View, there is an image. Fetch an image for a deco Preset.

    v: View
    d: dict
        Preset

    Return: Image or None
    """
    if d[ok.SWITCH] and d[ok.TYPE] == dc.IMAGE:
        e = d[ok.IMAGE_CHOICE]
        if e[ok.SWITCH]:
            return get_image_ref(e, v.image_source)


def assign_frame_image(v, frame_d, is_face):
    """
    Frame Over and Nail Polish may have a reference to an image.

    v: View
    frame_d: dict
        Frame type Preset

    is_face: bool
        Is True if the caller from the Face branch.

    Return: Image or None
    """
    def _nail_polish():
        _d = sub_frame_d[ok.WRW][ok.OVERLAY_NP]
        if _d[ok.TYPE] == ff.IMAGE:
            return get_image_ref(_d[ok.IMAGE_CHOICE], v.image_source)

    def _frame_over():
        _d = sub_frame_d[ok.WRW][ok.OVERLAY_FO]
        if _d[ok.TYPE] == ff.IMAGE:
            return get_image_ref(_d[ok.CIR][ok.IMAGE_CHOICE], v.image_source)

    # Face doesn't have a Frame option.
    if not is_face:
        if frame_d[ok.SWITCH]:
            option_k, sub_frame_d = get_option_list_choice(frame_d)
            p = {
                ek.FRAME_OVER: _frame_over, ek.NAIL_POLISH: _nail_polish
            }.get(option_k)
            if p:
                return p()


def assign_image(v, maya, q):
    """
    Assign Image for deco.

    v: View
    maya: Maya
    q: list
        [(r, c) or (r, c, x), ...]
    """
    d = maya.value_d
    per_d = maya.per_d
    per = d[ok.PER]

    for k in q:
        if k in per:
            this_maya = per_d[k]
            e = per[k]
            if k in d:
                maya.remove_ref(k)

        else:
            this_maya = maya
            e = d

        image = assign_deco_image(v, e)
        pre_image = this_maya.get_image(k)

        if image != pre_image:
            this_maya.is_matter = True
        this_maya.set_image(k, image)


def assign_mask_image(v, d):
    """
    Assign an image from a Mask Preset.

    v: View
    d: dict
        Mask Preset

    Return: Image or None
    """
    if d[ok.SWITCH] and d[ok.MASK_TYPE] == dc.IMAGE:
        e = d[ok.IBR][ok.IMAGE_CHOICE]
        if e[ok.SWITCH]:
            return get_image_ref(e, v.image_source)


def bag_vote(maya, d, k_path):
    """
    Extract vote from a Maya vote dict. Convert issue key
    found in a collected vote dict to Maya vote attribute.

    maya: Maya
        to receive issue vote

    d: dict
        Has relational vote.

    k_path: tuple
        (Option key, ...)
        Is a path to a relational structured vote dict.
    """
    for k in k_path:
        if k in d:
            d = d[k]
        else:
            d = {}
            break

    for i, a in d.items():
        if i in maya.issue_q and isinstance(a, dict):
            n = 'is_' + i
            vote = None

            for i1, b in a.items():
                if isinstance(b, bool):
                    if vote is None:
                        vote = b

                    else:
                        vote |= b
                    if vote:
                        break
            if vote is not None:
                setattr(maya, n, vote)
    take_great_vote(maya)


def check_overlay(v, maya):
    """
    Act on overlay issue change.

    v: View
    maya: Maya
    Return: layer or None
        with material
    """
    if maya.is_overlay:
        return check_layer(v, maya, 'matter', maya.do_matter)
    return maya.matter


def check_filler(v, maya):
    """
    Act on filler issue change.

    v: View
    maya: Maya
    Return: layer or None
        with material
    """
    if maya.is_filler:
        return check_layer(v, maya, 'matter', maya.do_matter)
    return maya.matter


def check_frame_mask_group(v, maya):
    """
    Make a group for Feather Step or Jagged Edge's mask.

    v: View
    maya: Maya
        Frame type

    Return: layer
        group
    """
    maya.group = make_frame_group(v, maya)

    if maya.group:
        z = maya.super_maya.group
        if z.parent != maya.group:
            pdb.gimp_image_reorder_item(v.j, z, maya.group, 0)
    return maya.group


def check_layer(v, maya, n, p):
    """
    Make a new layer and remove its old layer counterpart,
    if it exists. The layer is assumed to need an update.

    v: View
    maya: Maya

    n: string
        layer attribute descriptor

    p: function
        Call to make layer output.

    Return: layer or None
        with material
    """
    z = getattr(maya, n)

    if z:
        remove_maya_z(maya, n)
        v.is_back = True

    z = p(v, maya)

    if z:
        v.is_back = True
    return z


def check_matter(v, maya):
    """
    Act on matter issue change.

    v: View
    maya: Maya
    Return: layer or None
        with material
    """
    if maya.is_matter:
        return check_layer(v, maya, 'matter', maya.do_matter)
    return maya.matter


def check_metal_filler(v, maya):
    """
    Determine if a filler or an emboss issue has change. Make material if so.

    v: View
    maya: Maya
    Return: layer or None
        with material
    """
    if maya.is_filler or maya.is_emboss:
        return check_layer(v, maya, 'matter', maya.do_matter)
    return maya.matter


def check_mix(v, maya, issue, d, mode_k, opacity_k):
    """
    Determine a layer's mode or opacity change

    v: View
    maya: Maya

    issue: string
        Maya sub-attribute
        The string is prefix with 'is_' to create an attribute id.

    d: dict
        Preset
        Has option value.

    mode_k: string
        Option key for Mode

    option_k: string
        Option key for Opacity
    """
    z = maya.matter
    if z:
        if getattr(maya, 'is_' + issue):
            m = set_layer_attr(z, translate_mode(d[mode_k]), d[opacity_k])
            v.is_back += m
        else:
            if maya.is_mode:
                m = set_layer_mode(z, translate_mode(d[mode_k]))
                v.is_back += m
            if maya.is_opacity:
                f = d[opacity_k]
                if z.opacity != f:
                    z.opacity = f
                    v.is_back = True


def check_mix_basic(v, maya):
    """
    Determine the matter layer's mode or opacity change.

    v: View
    maya: Maya
    """
    check_mix(v, maya, 'matter', maya.value_d, ok.MODE, ok.OPACITY)


def check_mix_blur_behind(v, maya):
    """
    Determine a Blur Behind layer's mode or opacity change.

    v: View
    maya: Maya
    """
    check_mix(v, maya, 'blur_behind', maya.value_d, ok.MODE, ok.OPACITY)


def check_mix_overlay(v, maya):
    """
    Determine an overlay layer's mode or opacity change.

    v: View
    maya: Maya
    """
    check_mix(v, maya, 'overlay', maya.value_d, ok.MODE, ok.OPACITY)


def check_mix_noise(v, maya):
    """
    Determine a noise layer's mode or opacity change.

    v: View
    maya: Maya
    """
    check_mix(v, maya, 'noise', maya.value_d, ok.MODE, ok.OPACITY)


def check_mix_wrap(v, maya):
    """
    Determine a Metal Frame's matter layer's mode or opacity change.

    v: View
    maya: Maya
    """
    check_mix(
        v,
        maya,
        'matter',
        maya.value_d[ok.WRW][maya.wrap_k],
        ok.MODE,
        ok.OPACITY
    )


def check_mix_wrap_no_row(v, maya):
    """
    Determine a Metal Frame's matter layer's mode or opacity change.

    v: View
    maya: Maya
    """
    check_mix(
        v,
        maya,
        'matter',
        maya.value_d[maya.wrap_k],
        ok.MODE,
        ok.OPACITY
    )


def check_noise(v, maya):
    """
    Act on noise issue change.

    v: View
    maya: Maya
    Return: layer or None
        with material
    """
    if maya.is_noise:
        return check_layer(v, maya, 'matter', maya.do_matter)
    return maya.matter


def check_shadow(v, maya):
    """
    Act on shadow issue change.

    v: View
    maya: Maya
    Return: layer or None
        with material
    """
    if maya.is_shadow:
        return check_layer(v, maya, 'shadow', maya.do_matter)
    return maya.shadow


def format_image_name(j):
    """
    Get the name of an image, without its file
    extension, that is assigned to a cell.

    j: Image
        Has an image and layer name.

    Return: string
        an image name without its file type suffix
    """
    if j.layer_name:
        return j.layer_name

    if os.path.isfile(j.key):
        return os.path.basename(os.path.splitext(j.key)[0])
    return j.image_name


def get_sub_maya_vote(d, row_k, k):
    """
    From a hierarchical vote dict, find a sub-maya vote dict.

    d: dict
        Has issue vote.

    row_k: string
        Is a Row key where the option is found.

    k: string
        Is an Option key with accumulated vote.

    Return: dict
        Is the sub-Maya vote dict.
    """
    if row_k:
        a = d.get(row_k)
        if a:
            b = a.get(k)
            if b:
                return b

    a = d.get(k)

    if a:
        return a
    return {}


def init_per(maya, k):
    """
    Call before performing layer output during a View run.

    maya: Maya
        Per variety

    k: tuple
        (row, column) or (row, column, face)
        key to Per value
    """
    maya.value_d = maya.per_group.get_cell_value(k)
    n = ""

    for x, i in enumerate(k):
        n += str(i + 1)
        if x < len(k) - 1:
            n += "."
    maya.nav_k = maya.any_group.nav_k + (n,)


def make_frame_group(v, maya):
    """
    Make a Frame layer group. Is a sub-group of the super maya's layer group.

    v: View
    maya: Maya
    Return: layer group
        for frame
    """
    return make_sub_group(v, maya, " Frame")


def make_filler_group(v, maya):
    """
    Make a Filler layer group. Is a sub-group of the super Maya's layer group.

    v: View
    maya: Maya
    Return: layer group
        for frame
    """
    return make_sub_group(v, maya, " Filler")


def make_sub_group(v, maya, n):
    """
    Make a layer group. Is a sub-group of the super Maya's layer group.

    v: View
    maya: Maya
    n: string
        group name appendix

    Return: layer group
        for frame
    """
    parent = maya.cause.group.parent
    group = maya.group

    if not group:
        return make_layer_group(v.j, parent.name + n, parent=parent)
    return group


def make_source_key(d):
    """
    The source key is a signal key for notifying an image source sequence.

    d: dict
        Image Choice Preset

    Return: string
        signal key
    """
    k = d[ok.IMAGE_SOURCE]
    is_folder = k == ok.FOLDER
    key = (k,)

    if is_folder:
        if is_folder:
            key += (d[ok.FILTER],)
            if d[ok.RANDOM_ORDER]:
                key += (int(d[ok.SEED]),)
    return key


def on_global(maya, arg, option_k, issue=vo.MATTER):
    """
    Respond to a Global option change.

    maya: Maya
    arg: list
        [Plan vote, Work vote]

    option_k: string
        Option key
        Is responsible for the vote.

    issue: string
        Is altered by the vote.
    """
    d = maya.great_vote_d

    # Find the Cell Maya and set its 'per' vote to True.
    # The 'per' issue is voted on when a Per Cell Maya
    # has change, or in this case, when there is Global change.
    chain = maya

    while chain:
        if 'per' in chain.issue_q:
            # Found the Cell Maya, 'chain'.
            if not chain.is_per:
                a = chain.any_group

                a.accept_vote(1, option_k, vo.PER, True)
                chain.take_vote(a.vote_q[1])
                break
        chain = chain.super_maya if hasattr(chain, 'super_maya') else None

    if issue not in d:
        d[issue] = {}

    d[issue][option_k] = arg[maya.view_x]

    take_great_vote(maya)
    maya.any_group.changed(maya.view_x)


def prep_main_step(v, maya):
    """
    Undo a Per Cell type step. Init View's Deco container.

    v: View
    maya: Maya
    """
    v.deco = Deco()

    pdb.gimp_selection_none(v.j)
    if not maya.main_q:
        maya.die(v)


def retire_sub_maya(sub_maya):
    """
    Remove Maya layer output and its sub-Maya output.

    sub_maya: dict
        of sub-Maya
        {sub-Maya identifier: sub-Maya instance} -> {string: Maya}
    """
    for k, a in sub_maya.items():
        a.set_issue()
        retire_sub_maya(a.sub_maya)
        for i in a.order:
            if i:
                remove_maya_z(a, i)


def take_type_vote(v, d):
    """
    Border, Fringe, and Plaque have background
    sensitive types. If the background has changed
    with these types, then vote for change
    on that background-changed cause.

    v: View
    d: dict
        deco type Preset
        {Option key: value}

    Return: bool
        If true, then the matter layer has change.
    """
    if v.is_back and d[ok.TYPE] in (dc.AVERAGE_COLOR, dc.BACKDROP):
        return True
    return False


def take_great_vote(maya):
    """
    Set Maya issue from the great vote dict.

    maya: Maya
    """
    for i, a in maya.great_vote_d.items():
        if i in maya.issue_q and isinstance(a, dict):
            n = 'is_' + i
            vote = None

            for i1, b in a.items():
                if isinstance(b, bool):
                    if vote is None:
                        vote = b

                    else:
                        vote |= b
                    if vote:
                        break
            if vote is not None:
                vote |= getattr(maya, n)
                if vote:
                    setattr(maya, n, vote)


class Maya(object):
    """
    Is a base class for producing layer output. Is an object for Python 2.7.
    """

    def __init__(self, any_group, view_x, q, k_path=()):
        """
        any_group: AnyGroup
        q: list or None
            [(function, function output name)]

            Are function called in sequence. It's return value is
            assigned to an attribute defined by function output name.

        k_path: tuple or list
            (Option key, ...)
            [(Option key, ...), ...]

            A tuple key is a path to a sub-vote dict.
            The key is needed to find an option vote in a
            vote dict where some options are nested.

            A list of tuple have multiple key path where
            each path is to another vote dict collection
            perused when taking vote.
        """
        self.go = True
        self.handle_d = {}
        self.great_vote_d = {}
        self.k_path = k_path

        # None, (row, column), or (row, column, face), 'self.k'
        # Is a key for Maya and made for Model Goo synchronicity.
        self.k = None

        self.rect = None
        self.any_group = any_group
        self.view_x = view_x
        self.model = any_group.item.model
        self.value_d = any_group.value_d
        self.per_group = any_group.widget_d[ok.PER] \
            if ok.PER in any_group.widget_d \
            else None

        # Process during a View run. function list, 'self._make'.
        self._make = [i[0] for i in q]

        # Give name to function output layer
        # with an attribute list, 'self.order'.
        self.order = [i[1] for i in q]

        # Produce additional layer output with sub-Maya.
        self.sub_maya = {}

        # Are layer output reference (e.g. 'self.matter')
        for p, n in q:
            if n:
                setattr(self, n, None)

        if k_path != NO_VOTE:
            self.take_vote = self._mash_vote if isinstance(k_path, list) \
                else self._grab_vote

        else:
            self.take_vote = lambda *_q, **_d: None

        self.handle_d[The.cat.render.connect(
            si.CLOSE_VIEW_IMAGE, self.on_close_view_image
        )] = The.cat.render
        self.handle_d[The.power.connect(si.RESIZE, self.on_resize)] = The.power
        self.handle_d[
            The.power.connect(si.PANEL_CHANGE, self.on_panel_change)
        ] = The.power
        self.handle_d[
            self.any_group.connect(si.DISAPPEAR, self.on_disappear_group)
        ] = self.any_group
        if k_path != NO_VOTE:
            if hasattr(self, 'is_seeded'):
                self.handle_d[
                    The.power.connect(
                        si.GLOBAL_SEED, self.on_global_seed
                    )
                ] = The.power

            if hasattr(self, 'is_embossed'):
                self.handle_d[
                    The.power.connect(
                        si.GLOBAL_EMBOSS, self.on_global_emboss
                    )
                ] = The.power

            if self.vote_type == MAIN:
                self.handle_d[
                    any_group.booth.connect(
                        si.VOTE_CHANGE, self.on_vote_change
                    )
                ] = any_group.booth
                self.handle_d[
                    self.any_group.connect(si.SEQUENCE, self.on_sequence)
                ] = self.any_group
            elif self.vote_type == PER:
                # Per Cell Maya
                self.handle_d[
                    self.per_group.connect(
                        si.PER_CHANGE, self.on_per_change
                    )
                ] = self.per_group

    def _grab_vote(self, d):
        """
        Vote on issue given a vote dict.

        d: dict
            Has relational vote.
        """
        bag_vote(self, d, self.k_path)

    def _mash_vote(self, d):
        """
        Vote on issue given a vote dict with multiple vote sources.

        d: dict
            Has relational and nested vote.
        """
        for k_q in self.k_path:
            bag_vote(self, d, k_q)

    def die(self, v):
        """
        Remove its layer output.

        v: View
        """
        for i in self.order:
            if i:
                z = getattr(self, i)
                if z:
                    remove_z(z)
                    setattr(self, i, None)
                    v.is_back = True

        retire_sub_maya({None: self})
        self.set_issue()

    def on_close_view_image(self, _, arg):
        """
        Reset the Plan group reference as it was deleted.

        _: ViewImage
            Sent the signal.

        arg: not used
        """
        for i in self.order:
            if i:
                setattr(self, i, None)
        self.reset()

    def on_disappear(self, _, q):
        """
        The Maya is no longer used, so disconnect its subscription.

        q: tuple
            set of (row, column) or (row, column, face)
            Identify Maya with the view and cell index.
        """
        def _unplug(_maya):
            _maya.die(The.view)

            for _i, _a in _maya.handle_d.items():
                _a.disconnect(_i)
            _maya.handle_d = {}

        def _poof(_maya):
            for _, _maya in _maya.sub_maya.items():
                _unplug(_maya)
                _poof(_maya)
            _maya.sub_maya = {}

        for k in q:
            if k in self.per_d:
                maya = self.per_d[k]

                _unplug(maya)
                self.per_d.pop(k)

                # Remove the Maya's ability to take sub-vote.
                _poof(maya)
        pdb.gimp_displays_flush()

    def on_disappear_group(self, *_):
        """
        The Maya is no longer used, so disconnect its subscription.

        _: AnyGroup, arg
            (Sent the Signal., None)
        """
        self.die(The.view)

        for i, a in self.handle_d.items():
            a.disconnect(i)

        # Remove the Maya's ability to take sub-vote.
        self.per_d = self.sub_maya = self.handle_d = {}

        pdb.gimp_displays_flush()

    def on_global_emboss(self, _, arg):
        """
        Respond to a Global emboss option change.

        _: Power
            Sent the Signal.

        arg: list
            [Plan vote, Work vote]
        """
        on_global(self, arg, ok.IS_EMBOSS)

    def on_global_seed(self, _, arg):
        """
        Respond to a Global Seed option change.

        _: Power
            Sent the Signal.

        arg: list
            [Plan vote, Work vote]
        """
        on_global(self, arg, ok.IS_SEED)

    def on_panel_change(self, *_):
        """
        Check to see if the Maya is still active. If the Maya
        is shelved or has been deleted, then self-destruct.

        _: Power
            (Sent the Signal., arg: None)

        """
        k = self.any_group.nav_k
        if not The.helm.finds(k):
            self.die(The.view)

            for i, a in self.handle_d.items():
                a.disconnect(i)
            pdb.gimp_displays_flush()

    def on_per_change(self, _, arg):
        """
        Respond to change from its PerGroup.
        Create Per Cell Maya to mirror PerGroup cell addition.

        _: PerGroup
            Sent the Signal.

        arg: tuple
            ((row, column), vote dict, view index)
            The row and column work together as a cell index.
            The vote dict has vote for cell referenced by the cell index.
        """
        def _trickle(_d):
            """
            Recursively trickle down the vote through the sub-maya pyramid.

            _d: dict
                sub-Maya
                {sub-Maya key: Maya}
            """
            for _, _b in _d.items():
                _b.take_vote(d)
                _trickle(_b.sub_maya)

        k, d, x = arg
        if x == self.view_x and k == self.k:
            self.take_vote(d)
            _trickle(self.sub_maya)

    def on_resize(self, *_):
        """
        Cast a View image size change vote for both Plan and Work.

        *_: tuple
            (Power sent the signal., [Plan vote, Work vote])
        """
        on_global(self, [True, True], ok.IS_NEW)

    def on_sequence(self, *_):
        """
        Respond to change taking place in a Model sequence calculation.
        The 'matter' vote has already been added to the AnyGroup's vote
        dict, but now it needs to be scanned for True vote.

        _: tuple
            (AnyGroup, list)
            (Sent the Signal., [Plan vote, Work vote])
        """
        # Plan and Work, '2'
        for x in range(2):
            self.take_vote(self.any_group.vote_q[x])

    def on_vote_change(self, _, x):
        """
        Respond to VOTE_CHANGE Signal from its AnyGroup.
        Convert sub vote dict to Maya vote.

        x: int
            Plan or Work
            0 or 1
        """
        if x == self.view_x:
            self.take_vote(self.any_group.vote_q[x])

    def realize(self, v):
        """
        Create and delete layer output.

        v: View
        """
        if not self.go:
            self.die(v)
        else:
            for x, i in enumerate(self.order):
                a = self._make[x](v, self)
                if i:
                    setattr(self, i, a)

    def realize_vote(self, v):
        """
        Call to manage layer output and then reset the vote.

        v: View
        """
        self.realize(v)
        self.reset_issue()

    def reset(self):
        """Is called when the View image closes."""
        for i in self.issue_q:
            self.any_group.cast_vote(self.view_x, ok.IS_NEW, i, True)

    def reset_issue(self):
        """Indicate no change."""
        self.great_vote_d = {}
        for i in self.issue_q:
            setattr(self, 'is_' + i, False)

    def set_issue(self):
        """Set issue for self and sub-Maya to True."""
        for i in self.issue_q:
            setattr(self, 'is_' + i, True)
        for i, a in self.sub_maya.items():
            a.set_issue()


class Run(Maya):
    """Call functions during a View run."""

    def __init__(self, *q, **d):
        """
        q: tuple
            Maya spec

        d: dict
            Maya spec
        """
        Maya.__init__(self, *q, **d)

    def do(self, v):
        """
        Is an entry point for AnyGroup during a View run.

        v: View
        """
        self.prep(v)
        self.dig(v)


class ImageRecord:
    """Manage assigned image."""

    def __init__(self, *q, **d):
        """
        q: tuple
            Maya spec

        d: dict
            Maya spec
        """
        # The three dict that have the same
        # structure but different key configuration.
        # Canvas has a None key. Face and Facing have a face index.
        # {(r, c), (r, c, face), or None: Image or None}
        self._image_d = {}
        self._frame_image = {}
        self._mask_image = {}

    def get_facing_name(self, k):
        """
        Get a Facing image reference.

        k: tuple
            (row, column, face)
            cell index, Face index

        Return: string or None
            layer or image name
        """
        j = self.get_image(k)
        if j:
            return format_image_name(j)

    def get_frame_image(self, k):
        """
        Get a Frame image reference.

        k: tuple
            (row, column) or (row, column, face)
            Goo key from cell index

        Return: Image or None
            Assigned to the Frame Maya.
        """
        return self._frame_image.get(k)

    def get_image(self, k):
        """
        Fetch a Cell/Image or Canvas/Image reference.

        k: tuple or None
            Goo key
            Is None for Canvas.

        Return: Image or None
        """
        if hasattr(self, 'per_d'):
            if k in self.per_d:
                return self.per_d[k].get_image(k)
        if k in self._image_d:
            return self._image_d[k]

    def get_image_name(self, k):
        """
        Extract an image name from an Image reference.

        k: tuple
            (r, c), (r, c, x) or None

        Return: string or None
            layer or image name
        """
        j = self.get_image(k)
        if j:
            return format_image_name(j)

    def get_mask_image(self, k):
        """
        Get a mask image reference.

        k: tuple
            (row, column)
            Goo key; of int

        Return: the 'mask_image' attribute
        """
        return self._mask_image.get(k)

    def set_frame_image(self, k, j):
        """
        Set a Frame image reference.

        k: tuple or None
            (row, column), (row, column, face)
            zero-based indices

        j: Image
            to remember
        """
        self._frame_image[k] = j

    def set_image(self, k, j):
        """
        Set a Cell or Canvas branch image reference. Can
        be from deco type or an Image Preset's image choice.

        k: tuple
            (row, column), (row, column, face)
            zero-based indices

        j: Image
            Recall during a View run.
        """
        self._image_d[k] = j

    def set_mask_image(self, k, j):
        """
        The Maya's Preset has a mask reference.

        k: tuple or None
            (row, column), (row, column, face)
            zero-based indices

        j: Image
            to remember
        """
        self._mask_image[k] = j


class ImageRoll(ImageRecord, Run):
    """Manage assigned image."""

    def __init__(self, *q, **d):
        """
        q: tuple
            Maya spec

        d: dict
            Maya spec
        """
        ImageRecord.__init__(self, *q, **d)
        Run.__init__(self, *q, **d)


class PerRoute:
    """Factor CellRoute and FaceRoute"""

    def __init__(self, per_type):
        """
        per_type: Class
            Per Maya
        """
        self.per_type = per_type

        # Has Maya for each Per.
        # key, value -> {(r, c) or (r, c, x): Per Maya}
        self.per_d = {}

        # main step key
        self.nav_k = self.any_group.nav_k + (".",)

        self.handle_d[
            self.per_group.connect(si.DISAPPEAR, self.on_disappear)
        ] = self.per_group
        self.handle_d[
            self.per_group.connect(si.PER_CHANGE, self.on_per_d_change)
        ] = self.per_group

    def rig(self, v, q):
        """
        Process main and Per output for a step during a View run.

        v: View
        q: list
            Is a list of Per key for the Model.
        """
        d = self.per_d
        prep_main_step(v, self)
        self.bore(v)

        if self.is_per or v.is_back:
            for k in q:
                if k in d:
                    maya = d[k]

                    init_per(maya, k)
                    maya.bore(v)
                    maya.reset_issue()
        self.reset_issue()

    def on_per_d_change(self, _, arg):
        """
        Respond to change from the AnyGroup's PerGroup.
        Create Per Cell Maya to mirror PerGroup cell addition.

        _: PerGroup
            Sent the Signal.

        arg: tuple
            ((row, column) or (row, column, face), vote dict, view index)
            The row and column work together as a cell index and Goo key.
            The vote dict has vote for cell referenced by the cell index.
        """
        k, vote_d, x = arg
        if x == self.view_x:
            if k not in self.per_d:
                self.per_d[k] = self.per_type(self.any_group, k)


class CellRoute(PerRoute):
    """Has a list of Cell branch key."""

    def __init__(self, per_type):
        self.main_q = []
        PerRoute.__init__(self, per_type)

    def dig(self, v):
        self.main_q = get_main_q(self.value_d, self.model.cell_q)
        self.rig(v, self.model.cell_q)


class FaceRoute(PerRoute):
    """Has a list of Face branch key."""

    def __init__(self, per_type):
        self.main_q = []
        PerRoute.__init__(self, per_type)

    def dig(self, v):
        self.main_q = get_main_q(self.value_d, self.model.face_q)
        self.rig(v, self.model.face_q)


class Cell(CellRoute):
    """Use with deco step that secede Cell Margin."""

    def __init__(self, per_type):
        CellRoute.__init__(self, per_type)

    def prep(self, v):
        """
        Assign image. There is no Mask or Frame image.

        v: View
        """
        assign_image(v, self, self.model.cell_q)


class CanvasRoute:
    """Has Canvas branch function."""
    is_face = False
    vote_type = MAIN

    def __init__(self):
        self.k = None
        self.nav_k = self.any_group.nav_k + ("Main",)

    def dig(self, v):
        """
        Produce layer output.

        v: View
        """
        self.bore(v)
        self.reset_issue()


class Canvas(CanvasRoute):
    """Add Canvas image assignment."""

    def __init__(self):
        CanvasRoute.__init__(self)

    def prep(self, v):
        """
        For every View, there is an Image.

        v: View
        """
        image = assign_deco_image(v, self.value_d)

        if image != self.get_image(None):
            self.is_matter = True
        self.set_image(None, image)


class Plasma(ImageRoll):
    """Use with Border, Fringe and Plaque."""
    is_seeded = True

    def __init__(self, *q, **d):
        ImageRoll.__init__(self, *q, **d)

    def on_global_seed(self, _, arg):
        """
        The Global Random Seed option has changed.
        Check to see if the layer output is seed dependent.

        _: Power
            Sent the signal.

        arg: tuple
            (Plan vote, Work vote)
        """
        d = self.any_group.value_d if self.vote_type == MAIN else \
            self.per_group.get_cell_value(self.k)

        if d:
            if d[ok.TYPE] != dc.PLASMA:
                # no change
                arg = [False, False]
        super(Plasma, self).on_global_seed(_, arg)

    def on_back_change(self, _, x):
        """
        The layer background changed for the group.

        _: AnyGroup
            Sent the Signal.

        x: int
            Plan or Work index
        """
        if x == self.view_x:
            d = self.any_group.value_d if self.vote_type == MAIN else \
                self.per_group.get_cell_value(self.k)
            if d[ok.TYPE] in dc.BACKED:
                arg = [None, None]
                arg[x] = True
                on_global(self, arg, ok.IS_BACK)
