#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_key import Model as md
from roller_deco import ready_canvas_rect
from roller_polygon import get_bounds
from roller_fu import verify_layer_group
from roller_view_hub import add_text_layer, make_plan_text_layer
from roller_view_real import make_group


def draw_canvas_name(v, maya):
    """
    Draw image name for Canvas/Image.

    v: View
    maya: Maya
    Return: layer or None
        with the image name
    """
    ready_canvas_rect(v, maya, option=None)

    super_ = maya.super_maya
    j = v.j
    n = super_.get_image_name(None)
    if n:
        z = make_plan_text_layer(j, n, maya.font_size)
        x, y, w, h = maya.model.canvas_pocket.rect
        x = w / 2. + x - z.width / 2.
        y = h / 2. + y + z.height / 2.
        z.name = super_.group.name + " Name"

        add_text_layer(super_.group, z, x, y)
        return z


def draw_main_name(v, maya):
    """
    Draw an image name for a Cell branch image.

    v: View
    maya: Maya
    Return: layer or None
        with text
    """
    model = maya.model
    j = v.j
    super_ = maya.super_maya
    group = make_group(v, "Name", super_.group)
    font_size = maya.font_size

    for k in super_.main_q:
        a = model.get_pocket(k)
        x, y, w, h = a.rect
        n = super_.get_image_name(k)
        if n:
            z = make_plan_text_layer(j, n, font_size)
            x1 = w / 2. + x - z.width / 2.
            y1 = h / 2. + y + z.height / 2.

            if model.model_type == md.STACK:
                y1 = y1 + k[0] * z.height
            add_text_layer(group, z, x1, y1)
    return verify_layer_group(group)


def draw_cell_name(v, maya):
    """
    Draw an image name for a Cell branch image.

    v: View
    maya: Maya
    Return: layer or None
        with text
    """
    model = maya.model
    j = v.j
    super_ = maya.super_maya
    k = super_.k
    a = model.get_pocket(k)
    x, y, w, h = a.rect
    n = super_.get_image_name(k)
    if n:
        z = make_plan_text_layer(j, n, 13)
        x1 = w / 2. + x - z.width / 2.
        y1 = h / 2. + y + z.height / 2.

        if model.model_type == md.STACK:
            y1 = y1 + maya.k[0] * z.height
        return add_text_layer(super_.group, z, x1, y1)


def draw_face_per_name(v, maya):
    """
    Draw an image name for Plan Face/Per.

    v: View
    maya: Maya
    Return: layer or None
        with text
    """
    return produce_facing_per_name(v, maya, draw_face_name)


def draw_face_main_name(v, maya):
    """
    Draw Plan Face image name for for the main option settings.

    v: View
    maya: Maya
    Return: layer or None
        with text
    """
    super_ = maya.super_maya
    parent = super_.group
    group = make_group(v, "Name", parent)

    for k in super_.main_q:
        draw_face_name(maya, group, k)
    return verify_layer_group(group)


def draw_face_name(maya, group, k):
    """
    Draw image name for a Face.

    maya: Maya
    group: layer group
        Is where to place text layer.

    k: tuple
        (row, column, face)
        cell index
    """
    model = maya.model
    font_size = maya.font_size

    x, y, w, h = get_bounds(model.get_facing_foam(k))
    n = maya.super_maya.get_facing_name(k)
    if n:
        z = make_plan_text_layer(group.image, n, font_size)
        x1 = w / 2. + x - z.width / 2.
        y1 = h / 2. + y + z.height / 2.

        if model.model_type == md.STACK:
            y1 = y1 + k[0] * z.height
        add_text_layer(group, z, x1, y1)


def draw_facing_per_name(v, maya):
    """
    Draw an image name for Plan Facing/Per.

    v: View
    maya: Maya
    Return: layer or None
        with text
    """
    return produce_facing_per_name(v, maya, draw_facing_name)


def draw_facing_main_name(v, maya):
    """
    Draw an image name for Plan Face for the main option settings.

    v: View
    maya: Maya
    Return: layer or None
        with text
    """
    return produce_facing_main_name(v, maya, draw_facing_name)


def draw_facing_name(maya, group, k):
    """
    Draw image name for a Facing.

    maya: Maya
    group: layer group
        Is where to place text layer.

    k: tuple
        (row, column, None)
        cell index key
    """
    model = maya.model
    font_size = maya.font_size
    x, y, w, h = get_bounds(model.get_facing_foam(k))
    n = maya.super_maya.get_facing_name(k)
    if n:
        z = make_plan_text_layer(group.image, n, font_size)
        x1 = w / 2. + x - z.width / 2.
        y1 = h / 2. + y + z.height / 2.

        if model.model_type == md.STACK:
            y1 = y1 + k[0] * z.height
        add_text_layer(group, z, x1, y1)


def produce_facing_per_name(v, maya, p):
    """
    Draw an image name for Plan Facing/Per.

    v: View
    maya: Maya
    p: function
        Call to create cell Face/Facing name.

    Return: layer or None
        with text
    """
    super_ = maya.super_maya
    parent = super_.group
    group = make_group(v, "Name", parent)

    p(maya, group, *super_.k)
    return verify_layer_group(group)


def produce_facing_main_name(v, maya, p):
    """
    Draw an image name for Plan Face/Facing for the main option settings.

    v: View
    maya: Maya
    Return: layer or None
        with text
    """
    super_ = maya.super_maya
    parent = super_.group
    group = make_group(v, "Name", parent)

    for k in super_.main_q:
        p(maya, group, k)
    return verify_layer_group(group)
