#!/usr/bin/env python
# -*- coding: utf-8 -*-
from math import atan2
from roller_constant_for import Shape as sh, Signal as si
from roller_constant_key import Model as md, Option as ok
from roller_fu import clear_inverse_selection, select_rect
from roller_model_facing import Facing
from roller_model_goo import Goo, Map
from roller_model_grid import Grid
from roller_one_rect import Rect
from roller_one_the import The
from roller_polygon import calc_length, calc_rotated_point, shift_center

# foam key
CORNER = 'corner'
H_SIDE = 'h-side'
V_SIDE = 'v-side'

# {cell's foam type: (index to foam point in form polygon, ...)}
# Foam points are flag-ordered: topleft, top-right, bottom-left, bottom-right.
FOAM_D = {
    CORNER: (6, 4, 0, 2),
    H_SIDE: (0, 2, 6, 4),
    V_SIDE: (6, 0, 4, 2)
}


def calc_fill_rect(rect, is_right=False, is_bottom=False):
    """
    Calculate the scale of a cell rectangle that is to be rotated as a cover.
    The rotated rectangle is later clipped creating a fill for the Facing.
    Center the over-sized rectangle with Shift rectangle.

    rect: tuple
        (x, y, w, h)
        Is the Shift rectangle that is rotated.

    is_right, is_bottom: bool
        If True, then try to keep the fill rectangle within the View 'wip'.

    Return: tuple
        ((polygon array), (rectangle definition,), radius)
        The radius is the minimum span of the scaled cell rectangle.
    """
    x, y, w, h = rect
    diagonal = calc_length(x, y, x + w, y + h)
    wip = The.view.wip

    # grow factor, 'f'
    f = diagonal / min(w, h)
    w1, h1 = w * f, h * f

    x1 = x + w1
    y1 = y + h1

    if is_right:
        # Try to keep the fill rectangle 'x' within the bounds of 'wip.
        x2 = wip.w + wip.x
        if x1 > x2:
            w2 = x1 - x2
            x -= w2
            x1 -= w2

    if is_bottom:
        # Try to keep the fill rectangle 'y' within the bounds of 'wip.
        y2 = wip.h + wip.y
        if y1 > y2:
            h2 = y1 - y2
            y -= h2
            y1 -= h2
    return (x, y, x1, y, x1, y1, x, y1), (x, y, w1, h1)


def calc_sidewalk_cell(model):
    """
    Calc cell rectangle and form.

    model: Sidewalk

    Return: dict
        {(row, column0): [Plan vote, Work vote]}
    """
    vote_d = {}
    did_cell = model.past.did_cell
    row, column = model.division
    goo_d = model.goo_d
    x, y, canvas_w, canvas_h = model.canvas_pocket.rect
    w = canvas_w / column
    h = canvas_h / row

    # [intersect point]
    q_x, q_y = [], []

    # remainder total, 'f_x, f_y'
    f_x = f_y = .0

    # Add remainder to coordinate when close to one.
    for _ in range(column + 1):
        x, f = divmod(x, 1.)
        f_x += f

        if f_x >= .999:
            x += 1.
            f_x -= 1.

        q_x += [x]
        x += w

    for _ in range(row + 1):
        y, f = divmod(y, 1.)
        f_y += f

        if f_y >= .999:
            y += 1.
            f_y -= 1.

        q_y += [y]
        y += h

    for r_c in model.cell_q:
        r, c = r_c
        x, x1 = q_x[c], q_x[c + 1]
        y, y1 = q_y[r], q_y[r + 1]
        a = goo_d[r_c] = Goo(r_c)

        # Prevent round to zero with max 1.
        a.cell = a.merged = Rect(x, y, max(1., x1 - x), max(1., y1 - y))

        a.form = x, y, x1, y, x1, y1, x, y1
        vote_d[r_c] = did_cell(r_c)
    return vote_d


class Sidewalk(Grid, Facing):
    """
    Has a Canvas, Cell, and a Facing branch. Has two or
    more row and column arranged in a rectangular ring.
    """
    model_type = md.SIDEWALK

    def __init__(self, model_name):
        """
        model_name: string
            A key to identify the Model instance amongst other Models.
        """
        Grid.__init__(self, model_name)
        Facing.__init__(self)

        self.face_q = None
        self.cell_shape = sh.RECTANGLE
        self.baby.connect(si.CELL_MARGIN_CALC, self.update_sidewalk_facing)

    def calc_division(self, d):
        """
        Determine the row and column count of the cell grid.
        Call after Canvas pocket has been calculated.

        d: dict
            Cell Type Preset
            {Option key: value}
        """
        self.division = int(d[ok.ROW_COUNT]), int(d[ok.COLUMN_COUNT])
        self.baby.emit(si.DIVISION_CHANGE, self.division)

    def clip_facing(self, z, k):
        # Convert to (r, c, x) to (r, c).
        k = k[:2]

        select_rect(z.image, *self.goo_d[k].shift.rect)
        clear_inverse_selection(z)

    def init_cell_q(self, d):
        """
        Create a sorted list of valid cell index,
        'cell_q' -> [(row, column), ...].

        d: dict
            Cell Type Preset
        """
        row, column = self.division
        last_r = row - 1
        last_c = column - 1
        self.cell_q = []
        self.face_q = []

        for r in range(row):
            for c in range(column):
                if r == 0 or c == 0 or r == last_r or c == last_c:
                    self.cell_q += [(r, c)]
                    self.face_q += [(r, c, 0)]

    def init_model_cell(self, d):
        """
        Calculate cell, merged, and form value.

        d: dict
            Cell Type Preset
            {Option key: value}

        Return: list
            [Plan vote, Work vote]
            Each vote is a vote for change.
        """
        e = self.map_d = {}
        last_r, last_c = self.division[0] - 1, self.division[1] - 1
        e[(0, 0, 0)] = Topleft(d)
        e[(0, last_c, 0)] = TopRight(d)
        e[(last_r, 0, 0)] = BottomLeft(d)
        e[(last_r, last_c, 0)] = BottomRight(d)

        for i in range(1, last_r):
            e[(i, 0, 0)] = Left(d)
            e[(i, last_c, 0)] = Right(d)

        for i in range(1, last_c):
            e[(0, i, 0)] = Top(d)
            e[(last_r, i, 0)] = Bottom(d)

        self.goo_d = {}
        return calc_sidewalk_cell(self)

    def update_sidewalk_facing(self, *_):
        """Update Facing variables with Cell/Shift change."""
        for r_c in self.cell_q:
            self.map_d[r_c + (0,)].update_facing(self.goo_d[r_c])

    def update_type(self, arg):
        """
        Update Cell Type dependency.

        arg: tuple
            (Cell Type Preset, is sequence flag)
            {Option key: value}
            If the flag is True, then a chained-sequence
            of update processing is done immediately.
        """
        d, is_sequence = arg
        p = self.baby.feed if is_sequence else self.baby.give
        vote_d = super(Sidewalk, self).update_type(arg)

        self.adapt_missing_cell_step(is_sequence)
        p(si.CELL_RECT_CALC, vote_d)


class Form(Map):

    def __init__(self, d):
        """
        d: dict
            Cell Type Preset
        """
        Map.__init__(self, d, self.do_transform)
        self.angle = self.center_x = self.center_y = .0
        self.foam_form = None

    def calc_rotated_rect(self, q, radius):
        """
        Calculate rotated rectangle point.

        q: list
            [x, y, ...]
            Define a rectangle polygon.
            four points -> topleft, top-right, bottom-right, bottom-left

        radius: float
            Is the rectangle's half-diagonal.

        Return: list
            [x, y, ...]
            four points of float corresponding with the rectangle.
        """
        # rotated point polygon, 'q1'
        q1 = []

        for i in range(0, 8, 2):
            q1 += calc_rotated_point(
                self.center_x, self.center_y,
                atan2(q[i + 1], q[i]) + self.angle,
                radius
            )
        return q1

    def do_transform(self, v, maya):
        """
        Calculate the foam needed for a mold rectangle to transform into foam.

        v: View
        maya: Maya
        Return: list
            [x, y, ...]
            Are four x, y coordinates arranged in flag-order for
            the 'pdb.gimp_item_transform_perspective' function.
            Flag-order is topleft, top-right, bottom-left, bottom-right.
        """
        x, y, w, h = v.pot.mold.rect
        x1, y1 = x + w, y + h
        radius = calc_length(x, y, x1, y1) / 2.
        foam_mold = shift_center((x, y, x1, y, x1, y1, x, y1), *v.pot.mold.center())

        # Calculate polygon point that is ready for selection.
        return self.pour_foam(
            self.calc_rotated_rect(foam_mold, radius)
        )

    def pour_foam(self, q):
        """
        q: iterable
            form polygon

        Return: list
            [x, y, ...] in flag-order
            foam
        """
        # foam, 'q2'
        # [x, y, ...]
        q2 = []

        if self.is_reverse and self.inward:
            q1 = FOAM_D[self.foam_k][::-1]

        elif not self.is_reverse and not self.inward:
            q1 = FOAM_D[self.foam_k][::-1]

        else:
            q1 = FOAM_D[self.foam_k]

        for x in q1:
            q2 += [q[x], q[x + 1]]
        return q2


class Bottom(Form):
    """
    Define property for a bottom row cell that
    is not a corner in the Sidewalk grid.
    """
    foam_k = H_SIDE
    is_reverse = True

    def __init__(self, d):
        """
        d: dict
            Cell Type Preset
        """
        Form.__init__(self, d)
        # 180 degrees
        self.angle = 3.141592653

    def update_facing(self, goo):
        self.merged.rect = goo.shift.rect
        self.form = self.shape = goo.plaque
        self.foam = self.pour_foam(self.form)


class BottomLeft(Form):
    """Define property for the bottom-left corner cell in the Sidewalk grid."""
    foam_k = CORNER
    is_reverse = True

    def __init__(self, d):
        """
        d: dict
            Cell Type Preset
        """
        Form.__init__(self, d)

    def update_facing(self, goo):
        """
        A Facing is dependent on the Shift
        rectangle, so respond to change in it.

        goo: Goo
        """
        self.form, self.merged.rect = calc_fill_rect(
            goo.shift.rect, is_bottom=True
        )
        self.center_x, self.center_y = goo.shift.center()
        self.shape = goo.plaque

        # 180 degrees, '3.14'
        self.angle = -atan2(goo.shift.h, goo.shift.w) + 3.141592653

        # Make point usable for atan2.
        self.foam_form = shift_center(self.form, *self.merged.center())

        x, y, w, h = self.merged.rect
        radius = calc_length(x, y, x + w, y + h) / 2.

        # Calculate polygon point that is ready for selection.
        self.foam = self.pour_foam(
            self.calc_rotated_rect(self.foam_form, radius)
        )


class BottomRight(Form):
    """
    Define property for the bottom-right corner cell in the Sidewalk grid.
    """
    foam_k = CORNER
    is_reverse = False

    def __init__(self, d):
        """
        d: dict
            Cell Type Preset
        """
        Form.__init__(self, d)

    def update_facing(self, goo):
        """
        A Facing is dependent on the Shift
        rectangle, so respond to change in it.

        goo: Goo
        """
        self.form, self.merged.rect = calc_fill_rect(
            goo.shift.rect, is_bottom=True, is_right=True
        )
        self.center_x, self.center_y = goo.shift.center()
        self.shape = goo.plaque

        # 180 degrees, '3.14'
        self.angle = atan2(goo.shift.h, goo.shift.w) + 3.141592653

        # Make point usable for atan2.
        self.foam_form = shift_center(self.form, *self.merged.center())

        x, y, w, h = self.merged.rect
        radius = calc_length(x, y, x + w, y + h) / 2.

        # Calculate polygon point that is ready for selection.
        self.foam = self.pour_foam(
            self.calc_rotated_rect(self.foam_form, radius)
        )


class Left(Form):
    """
    Define property for a left column cell that
    is not a corner in the Sidewalk grid.
    """
    foam_k = V_SIDE
    is_reverse = False

    def __init__(self, d):
        """
        d: dict
            Cell Type Preset
        """
        Form.__init__(self, d)

        # 90 degrees, '1.57'
        self.angle = 1.570796326

    def update_facing(self, goo):
        """
        A Facing is dependent on the Shift
        rectangle, so respond to change in it.

        goo: Goo
        """
        # The width and height exchange places as the cell is on its side.
        x, y, w, h = goo.shift.rect

        self.merged = Rect(x, y, h, w)
        self.form = self.shape = goo.plaque
        self.foam = self.pour_foam(self.form)


class Right(Form):
    """
    Define property for a right column cell that
    is not a corner in the Sidewalk grid.
    """
    foam_k = V_SIDE
    is_reverse = True

    def __init__(self, d):
        """
        d: dict
            Cell Type Preset
        """
        Form.__init__(self, d)

        # 270 degree, '4.71'
        self.angle = 4.71238898

    def update_facing(self, goo):
        """
        A Facing is dependent on the Shift
        rectangle, so respond to change in it.

        goo: Goo
        """
        # The width and height exchange places as the cell is on its side.
        x, y, w, h = goo.shift.rect

        self.merged = Rect(x, y, h, w)
        self.form = self.shape = goo.plaque
        self.foam = self.pour_foam(self.form)


class Top(Form):
    """
    Define property for a top row cell that
    is not a corner in the Sidewalk grid.
    """
    foam_k = H_SIDE
    is_reverse = False

    def __init__(self, d):
        """
        d: dict
            Cell Type Preset
        """
        Form.__init__(self, d)

    def update_facing(self, goo):
        """
        A Facing is dependent on the Shift
        rectangle, so respond to change in it.

        goo: Goo
        """
        self.merged.rect = goo.shift.rect
        self.form = self.shape = goo.plaque
        self.foam = self.pour_foam(self.form)


class Topleft(Form):
    """Define property for the topleft corner cell in the Sidewalk grid."""
    foam_k = CORNER
    is_reverse = False

    def __init__(self, d):
        """
        d: dict
            Cell Type Preset
        """
        Form.__init__(self, d)

    def update_facing(self, goo):
        """
        A Facing is dependent on the Shift
        rectangle, so respond to change in it.

        goo: Goo
        """
        self.form, self.merged.rect = calc_fill_rect(goo.shift.rect)
        self.center_x, self.center_y = goo.shift.center()
        self.shape = goo.plaque
        self.angle = atan2(goo.shift.h, goo.shift.w)

        # Make point usable for atan2.
        self.foam_form = shift_center(self.form, *self.merged.center())

        x, y, w, h = self.merged.rect
        radius = calc_length(x, y, x + w, y + h) / 2.

        # Calculate polygon point that is ready for selection.
        self.foam = self.pour_foam(
            self.calc_rotated_rect(self.foam_form, radius)
        )


class TopRight(Form):
    """Define property for the top-right corner cell in the Sidewalk grid."""
    foam_k = CORNER
    is_reverse = True

    def __init__(self, d):
        """
        d: dict
            Cell Type Preset
        """
        Form.__init__(self, d)

    def update_facing(self, goo):
        """
        A Facing is dependent on the Shift
        rectangle, so respond to change in it.

        goo: Goo
        """
        self.form, self.merged.rect = calc_fill_rect(
            goo.shift.rect, is_right=True
        )
        self.center_x, self.center_y = goo.shift.center()
        self.shape = goo.plaque
        self.angle = -atan2(goo.shift.h, goo.shift.w)

        # Make point usable for atan2.
        self.foam_form = shift_center(self.form, *self.merged.center())

        x, y, w, h = self.merged.rect
        radius = calc_length(x, y, x + w, y + h) / 2.

        # Calculate polygon point that is ready for selection.
        self.foam = self.pour_foam(
            self.calc_rotated_rect(self.foam_form, radius)
        )
